#!/usr/bin/env bash

# Example usage:
#   echo "select distinct type, count(*) from node group by 1 order by 2 desc;" > node_counts.sql
#   ./query_all.sh node_counts.sql live | tee node_counts.txt

UPSTREAM='5161e51a-0e4a-414c-974e-8565094b76b1';

# Check command line arguments
if [[ -z $1 ]]; then
  echo; echo "Usage: $0 <query_file> [environment]";
  echo; echo;
  exit 1;
elif [[ ! -f $1 || ! -r $1 ]]; then
  echo; echo "SQL file: The filename '$1' doesn't exist, isn't readable, or isn't an ordinary file.";
  echo; echo;
  exit 1;
fi


# Default to 'dev' environment if none was specified
if [[ -z $2 ]]; then
  >&2 echo; >&2 echo "No environment specified, assuming dev.."; >&2 echo;
  ENV='dev';
else
  ENV=$2;
fi


# We use a second "inner script" because we prefer to use GNU 'parallel' for speed if it's available, and I can't
# figure out how to run a multi-command subprocess with parallel without putting it in a second script.
INNER_SCRIPT='./query_site.sh';

# Check $INNER_SCRIPT exists and is executable
if [[ ! -f $INNER_SCRIPT || ! -x $INNER_SCRIPT ]]; then
  echo; echo "'$INNER_SCRIPT' isn't an executable script. Stopping.";
  echo; echo;
  exit 1;
fi


# Check if GNU parallel
USE_PARALLEL=0;
PARALLEL_PATH=$(which parallel);   # 'parallel' command exists
# >&2 echo "'parallel' path: $PARALLEL_PATH" > /dev/stderr;

if [[ ! -z $PARALLEL_PATH ]]; then
  PARALLEL_VERSION=$($PARALLEL_PATH --version | head -1); # get first line of version
  # >&2 echo "'parallel' version: $PARALLEL_VERSION" > /dev/stderr;
  if [[ $PARALLEL_VERSION == *"GNU parallel"* ]]; then
    # >&2 echo "GNU parallel found: $PARALLEL_PATH" > /dev/stderr;
    USE_PARALLEL=1;   # it really is GNU parallel.  we like it.
  fi
fi


# ALL sites
>&2 echo -n "Getting site list... ";

SITES=$(terminus site:list --upstream=$UPSTREAM --fields=name --format=string | sort);
# SITES="comm-d8 comparativestudies-d8 complexity-d8 cstw-d8";
>&2 echo $SITES; >&2 echo;

# Non-sandbox sites only
# SITES=$(terminus site:list --upstream=$UPSTREAM --fields=name --format=string --filter=plan_name!=Sandbox | sort);


# Determine whether to generate backups
QUERY_ALL_GENERATE_BACKUPS=0;
if [[ -z $QUERY_ALL_SKIP_BACKUPS && "$ENV" == "live" ]]; then
  QUERY_ALL_GENERATE_BACKUPS=1;
fi
export QUERY_ALL_GENERATE_BACKUPS=1;


# Run it
echo > /dev/stderr;
if [[ $USE_PARALLEL -gt 0 ]]; then
  # With parallel
  parallel --delay 0.1 --tag -j 36 --line-buffer "$INNER_SCRIPT {}.$ENV $1" ::: $SITES
else
  # We're going to use the inner script even without parallel
  # because DRY and I can't be bothered to update the logic in 2 places.

  for SITE_NAME in $SITES; do
    echo "=== $SITE_NAME ===";
    $INNER_SCRIPT $SITE_NAME.$ENV $1;
    echo;echo;
  done
fi

